/*
 * enhashctf: Take the tokens in a ctf file and add them to a db file as a
 * set of token tuples which allow fast lookup of matching runs of tokens
 * between multiple ctf files.
 * Copyright (c) Warren Toomey, under the GPL license.
 * 
 * $Revision: 1.23 $
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "tokens.h"
#include "dbfiles.h"

#undef DUMMY_PRINT

/* Create and save a hash/node into the database */
void save_tuple(uint8_t *hash, uint16_t *idhash, Ctfhashnode node)
{
  int i;

  /* Calculate the XOR of all the id hashes */
  uint16_t idxor= idhash[0];
  for (i=1; i<TUPLE_SIZE;i++)
    idxor ^= idhash[i];

#ifdef DUMMY_PRINT
  for (i=0; i<TUPLE_SIZE;i++)
    printf("%02x", hash[i]);
  printf(":%04x\n  offset %04x name %04x file %02d lines %03d to %03d\n",
	idxor, node.offset, node.name_offset, node.fileid,
	node.line_number, node.last_line_number);
  printf("  ");
  for (i=0; i<TUPLE_SIZE;i++)
    printf("%04x ", idhash[i]);
  printf("\n");
#else

  /* Insert the idxor into the tuple array */
  hash[TUPLE_SIZE]=   idxor >> 8;
  hash[TUPLE_SIZE+1]= idxor & 0xff;

  add_ctfnode(hash, &node);
#endif
}

int main(int argc, char *argv[])
{
  /* Note: tuple[] is 2 bytes wider so we can insert the XOR'd idhash later */
  uint8_t tuple[TUPLE_SIZE+sizeof(uint16_t)]; 	/* tuple of tokens found */
  int token_offset[TUPLE_SIZE];	/* Offset of last TUPLE_SIZE tokens */
  int token_linenum[TUPLE_SIZE];/* Line number of last TUPLE_SIZE tokens */
  uint16_t idhash[TUPLE_SIZE];	/* Id value for each of the tokens in tuple */
  int i;
  unsigned int ch;		/* Incoming token */
  Ctfhashnode node;		/* Node to insert into the database */
  int token_count=0;		/* # of valid tokens found since filename */
  uint32_t old_offset;		/* Offset of current token */
  uint32_t offset=0;		/* Offset of next token */
  uint32_t name_offset=0;	/* Offset of last file name */
  int line_number=0;		/* Current line number */
  int index = 0;		/* Index into the lists (modulo TUPLE_SIZE) */
  int fileid;			/* New file id of this file */
  uint32_t idval;		/* Id value for token fetched */
  char **name;			/* Pointer to each filename in the CTF file */
  int add_it=1;			/* If 1, the tuple is added */
  FILE *zin;

  /* Check the input arguments */
  if ((argc != 2) && (argc != 3)) {
    fprintf(stderr, "Usage: enhashctf cft_file [start_pattern]\n"); exit(1);
  }

  /* Open up the named input file */
  zin= ctfopen(argv[1]);

  /* Initialise the name and offsets to after the header */
  name = (char **)malloc(sizeof(char *));
  offset = ftell(zin);
  old_offset = offset;

  /* Add the filename to the database and get the fileid back */
  fileid = add_fileidname(argv[1]);

  /* While there is a token to be found */
  while ((ch = get_token(zin, &offset, &idval, name)) != 0) {
    switch (ch) {
    case FILENAME:
      /*
       * We found a new filename. Clear the tuple,
       * save the offset and reset some counters.
       */
      name_offset = old_offset;
      for (i=0; i< TUPLE_SIZE; i++) tuple[i]=0;
      token_count = 0; line_number = 1;
      index = 0; add_it=1;

      /* Check and set the dont_add_it flag if we don't want it */
      if (argc == 3) {
        char *result = strstr(*name, argv[2]);
        if (result != *name) add_it=0;
      }
      if (add_it)
        printf("%s\n", *name);
      free(*name);
      break;

    case LINE:
      line_number++;
      break;

    default:
      /* Add the token and id value to the tuple */
      for (i = 0; i < TUPLE_SIZE - 1; i++) {
        tuple[i]  = tuple[i + 1];
        idhash[i] = idhash[i + 1];
      }

      /* Only add the id hash if not an IDENTIFIER. Otherwise, use 0 */
      tuple[TUPLE_SIZE-1]  = ch;
      idhash[TUPLE_SIZE-1] = (ch == IDENTIFIER) ? 0 : idval & 0xffff;
      token_count++;

      /* Before we get the next token, save this one's offset and line number */
      token_offset[index]  = old_offset;
      token_linenum[index] = line_number;
      index = (index + 1) % TUPLE_SIZE;

      /* Save any existing tuple */
      if (token_count >= TUPLE_SIZE) {
	node.offset = token_offset[index];
	node.line_number = token_linenum[index];
	node.last_line_number = line_number;
	node.name_offset = name_offset;
	node.fileid = fileid;
        if (add_it)
	  save_tuple(tuple, idhash, node);
      }
    }
    old_offset = offset;
  }
  close_ctfnode();
  exit(0);
}
