/*
 * gdbmfiles: Functions that deal with gdbm files and the nodes contained in
 * them.
 * Copyright (c) Warren Toomey, under the GPL license.
 * 
 * $Revision: 1.17 $
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <gdbm.h>
#include "dbfiles.h"

/* Static variables, used to hold state here. */

static GDBM_FILE ctf_dbf = NULL;	/* File handle for ctfnode.db file */

/*** Fileid functions ***/

/* Given an open gdbm file handle, return the
 * highest fileid in the database. Also check
 * the existence of the given filename.
 */
static int get_highest_fileid(GDBM_FILE dbf, char *name)
{
  datum key;
  int *id, fileid = 0;		/* Highest fileid found */
  /* Find all the keys in the database: they are the file-ids */
  key = gdbm_firstkey(dbf);
  while (key.dptr) {
    id = (int *)key.dptr;

    /* Ensure the ctf filename is not in there already */
    if (name != NULL) {
      datum val = gdbm_fetch(dbf, key);
      if (val.dptr && (!strcmp(val.dptr, name))) {
        fprintf(stderr,
	      "ctf file %s is already in the database as id %d, exiting\n",
	      name, *id);
        gdbm_close(dbf);
        exit(1);
      }
    }

    /* Find and update the highest fileid found */
    /* printf("Found fileid %d: %s\n", *id, val.dptr); */
    if (*id > fileid)
      fileid = *id;
    key = gdbm_nextkey(dbf, key);
  }
  return(fileid);
}

/*
 * Given a filename, add it to the fileid database and return the fileid. On
 * failure, return the value 0.
 */
int add_fileidname(char *name)
{
  GDBM_FILE dbf;		/* File handle */
  int fileid;			/* New fileid to add to the database */
  datum key;			/* Key retrieved and added */
  datum val;			/* Value added */

  /* Reject NULL or empty strings */
  if (name == NULL || strlen(name) == 0)
    return (0);

  /* Open up the database for read/write */
  dbf = gdbm_open(FILEID_DB, 0, GDBM_WRCREAT, 0666, NULL);
  if (dbf == NULL) {
    fprintf(stderr, "Unable to open the %s gdbm database: %s\n", FILEID_DB,
	    gdbm_strerror(gdbm_errno));
    exit(1);
  }

  /* Now write a new fileid to the database */
  fileid= get_highest_fileid(dbf,name) + 1;
  key.dptr = (char *)&fileid;
  key.dsize = sizeof(fileid);
  val.dptr = name;
  val.dsize = strlen(name)+1;	/* Ensure we add the \0 */

  if (gdbm_store(dbf, key, val, GDBM_INSERT) != 0) {
    fprintf(stderr, "Unable to insert %d:%s into the %s gdbm database: %s\n",
	    fileid, name, FILEID_DB, gdbm_strerror(gdbm_errno));
    gdbm_close(dbf);
    exit(1);
  }

  /* Close the database and return the fileid */
  gdbm_close(dbf);
  return (fileid);
}

/* Return a pointer to an array of filenames in the
 * database ordered by their fileid, and the number
 * of fileids (including the unused zero one).
 */
int get_fileidnames(char ***outarray)
{
  GDBM_FILE dbf;		/* File handle */
  char **namearray;
  datum key,val;
  int fileid,numfileids;

  if (outarray ==NULL) return(0);

  /* Open up the database for read/write */
  dbf = gdbm_open(FILEID_DB, 0, GDBM_READER, 0666, NULL);
  if (dbf == NULL) {
    fprintf(stderr, "Unable to open the %s gdbm database: %s\n", FILEID_DB,
	    gdbm_strerror(gdbm_errno));
    exit(1);
  }

  /* Get the number of fileids in the database */
  numfileids= get_highest_fileid(dbf,NULL) + 1;

  /* Allocate an array to hold all the file names */
  namearray= (char **)malloc(numfileids * sizeof(char *));

  /* Now read in each of the names */
  key.dptr = (char *)&fileid;
  key.dsize = sizeof(fileid);
  for (fileid=1; fileid < numfileids; fileid++) {
    val= gdbm_fetch(dbf, key);
    namearray[fileid]= val.dptr;
  }
  *outarray= namearray;
  
  gdbm_close(dbf);
  return(numfileids);
}

/*
 * Open the ctfnode.db with the given parameter, and save the filehandle in
 * the global variable. On failure, print out an error message and kill the
 * program.
 */
static void open_ctfnodedb(int mode)
{
  if (ctf_dbf != NULL) {
    fprintf(stderr, "Cannot re-open already open file %s, exiting\n",
	    CTFNODE_DB);
    exit(1);
  }

  ctf_dbf = gdbm_open(CTFNODE_DB, 0, mode, 0666, NULL);
  if (ctf_dbf == NULL) {
    fprintf(stderr, "Unable to open the %s gdbm database: %s\n", CTFNODE_DB,
	    gdbm_strerror(gdbm_errno));
    exit(1);
  }
}

/*
 * Create and add a new ctf node into the ctfnode file. We need to check if
 * the key is already in use. If so, read the existing data, delete the old
 * node, augment the new node and write it back. Returns 1 if OK, 0 if fail.
 */
int add_ctfnode(uint8_t * tuple, Ctfhashnode * value)
{
  datum oldval;			/* Any old value for key */
  datum key, val;
  char *newdptr = NULL, *insertposn;

  /* Inputs must exist */
  if ((tuple == NULL) || (value == NULL)) 
    return(0);

  key.dptr= (char *)tuple;
  key.dsize= TUPLE_SIZE+sizeof(uint16_t);
  val.dptr= (char *)value;
  val.dsize= sizeof(Ctfhashnode);

  /* Open the ctfnode.db file */
  if (ctf_dbf == NULL)
    open_ctfnodedb(GDBM_WRCREAT);

  /* Get any old value */
  oldval = gdbm_fetch(ctf_dbf, key);
  if (oldval.dptr) {
#if 0
    int i;
    printf("Found existing val for key ");
    for (i=0; i< TUPLE_SIZE; i++) printf("%02x", key.dptr[i]);
    printf(", size %d\n", oldval.dsize);
#endif

    /* Realloc and append the val's Ctfhashnode to the end of oldval's data */
    newdptr = realloc(oldval.dptr, oldval.dsize + sizeof(Ctfhashnode));
    if (newdptr == NULL)
      return (0);

    insertposn = newdptr + oldval.dsize;
    memcpy(insertposn, val.dptr, sizeof(Ctfhashnode));
    val.dptr = newdptr;
    val.dsize = oldval.dsize + sizeof(Ctfhashnode);
  } else {
#if 0
    int i;
    printf("Storing first val for key ");
    for (i=0; i< TUPLE_SIZE; i++) printf("%02x", key.dptr[i]);
    printf("\n");
#endif
  }

  int result = gdbm_store(ctf_dbf, key, val, GDBM_REPLACE);
  if (newdptr) free(newdptr);
  if (result != 0) return (0);
  return (1);
}


/* Close the close_ctfnode, and shrink it */
void close_ctfnode(void)
{
  if (ctf_dbf) {
    printf("Reorganising %s\n", CTFNODE_DB);
    gdbm_reorganize(ctf_dbf);
    gdbm_close(ctf_dbf);
  }
}

/* Get the first key in the database and its value.
 * Size is # Ctfhashnodes in the value.
 * Returns 1 if OK, 0 otherwise.
 */
int get_firstkey(uint8_t ** tuple, Ctfhashnode ** value, int *size)
{
  datum key, val;

  /* Inputs must exist */
  if ((tuple == NULL) || (value == NULL)) 
    return(0);

  /* Open the ctfnode.db file */
  if (ctf_dbf == NULL)
    open_ctfnodedb(GDBM_WRCREAT);

  key = gdbm_firstkey(ctf_dbf);
  if (key.dptr==NULL) return(0);

  val = gdbm_fetch(ctf_dbf, key);
  if (val.dptr==NULL) return(0);

  *tuple= (uint8_t *)key.dptr;
  *value= (Ctfhashnode *)val.dptr;
  *size= val.dsize / sizeof(Ctfhashnode);
  return(1);
}

/* Get the next key in the database and its value.
 * Size is # Ctfhashnodes in the value.
 * Returns 1 if OK, 0 otherwise.
 */
int get_nextkey(uint8_t *oldtuple, uint8_t ** tuple, Ctfhashnode ** value,
					int *size)
{
  datum oldkey, key, val;

  /* Inputs must exist */
  if ((oldtuple == NULL) || (tuple == NULL) || (value == NULL)) 
    return(0);

  /* Open the ctfnode.db file */
  if (ctf_dbf == NULL)
    open_ctfnodedb(GDBM_WRCREAT);

  oldkey.dptr= (char *)oldtuple;
  oldkey.dsize= TUPLE_SIZE+sizeof(uint16_t);

  key = gdbm_nextkey(ctf_dbf, oldkey);
  if (key.dptr==NULL) return(0);

  val = gdbm_fetch(ctf_dbf, key);
  if (val.dptr==NULL) return(0);

  *tuple= (uint8_t *)key.dptr;
  *value= (Ctfhashnode *)val.dptr;
  *size= val.dsize / sizeof(Ctfhashnode);
  return(1);
}
